<!DOCTYPE html>
<meta charset="utf-8">
<title>0 A.D. report service</title>
<style>
body {
    font-size: 12px;
    font-family: sans-serif;
}
</style>

<p>This site collects opt-in automatic feedback from players of <a href="http://play0ad.com/">0 A.D.</a>

<p>Published data:
<a href="{% url userreport.views.report_opengl_index %}">OpenGL capabilities</a>.
<a href="{% url userreport.views.report_cpu %}">CPU features</a>.
