/** 
 *	 tablefilter v0.0.11 by Max Guglielmi 
 *	 build date: 2015-11-21T07:50:21.705Z 
 *	 MIT License  
 */ 
span.expClpFlt a.btnExpClpFlt{width:35px;height:35px;display:inline-block;}span.expClpFlt a.btnExpClpFlt:hover{background-color:#f4f4f4}span.expClpFlt img{padding:8px 11px 11px 11px}