/* Copyright (C) 2018 Wildfire Games.
 * This file is part of 0 A.D.
 *
 * 0 A.D. is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * 0 A.D. is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 0 A.D.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef INCLUDED_JSI_USERREPORT
#define INCLUDED_JSI_USERREPORT

#include "scriptinterface/ScriptInterface.h"

#include <string>

namespace JSI_UserReport
{
	bool IsUserReportEnabled(ScriptInterface::CxPrivate* pCxPrivate);
	void SetUserReportEnabled(ScriptInterface::CxPrivate* pCxPrivate, bool enabled);
	std::string GetUserReportStatus(ScriptInterface::CxPrivate* pCxPrivate);
	std::string GetUserReportLogPath(ScriptInterface::CxPrivate* pCxPrivate);
	std::string GetUserReportConfigPath(ScriptInterface::CxPrivate* pCxPrivate);

	void RegisterScriptFunctions(const ScriptInterface& ScriptInterface);
}

#endif // INCLUDED_JSI_USERREPORT
