return {
	"_preload.lua",
	"vstudio.lua",
	"vs2005.lua",
	"vs2008.lua",
	"vs200x_vcproj.lua",
	"vs200x_vcproj_user.lua",
	"vs2005_solution.lua",
	"vs2005_csproj.lua",
	"vs2005_csproj_user.lua",
	"vs2010.lua",
	"vs2010_nuget.lua",
	"vs2010_vcxproj.lua",
	"vs2010_vcxproj_user.lua",
	"vs2010_vcxproj_filters.lua",
	"vs2010_rules_props.lua",
	"vs2010_rules_targets.lua",
	"vs2010_rules_xml.lua",
	"vs2012.lua",
	"vs2013.lua",
	"vs2015.lua",
	"vs2017.lua",
}
