// -*- C++ -*-
// See UserTraits.h

#define CXXTEST_USER_VALUE_TRAITS
#include <cxxtest/ErrorPrinter.h>

int main()
{
    return CxxTest::ErrorPrinter().run();
}

// The CxxTest "world"
<CxxTest world>


//
// Local Variables:
// compile-command: "perl test.pl"
// End:
//
