/* Copyright (C) 2015 Wildfire Games.
 * This file is part of 0 A.D.
 *
 * 0 A.D. is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * 0 A.D. is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 0 A.D.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
GUI Inclusion file

--Overview--

	Include this file and it will include the whole GUI.

--More info--

	Check TDD for GUI Engine Documentation

*/


#ifndef INCLUDED_GUI
#define INCLUDED_GUI

#include <map>
#include <stddef.h>
#include <string>
#include <vector>

#include "ps/CStr.h"

#include "CGUIList.h"
#include "CGUISeries.h"
#include "GUIbase.h"
#include "GUItext.h"
#include "GUIutil.h"
#include "IGUIButtonBehavior.h"
#include "IGUIObject.h"
#include "IGUIScrollBarOwner.h" // Required by IGUIScrollBar
#include "IGUIScrollBar.h"
#include "IGUITextOwner.h"

#endif // INCLUDED_GUI
