/*
	Copyright (C) 2005-2007 Feeling Software Inc.
	Portions of the code are:
	Copyright (C) 2005-2007 Sony Computer Entertainment America
	
	MIT License: http://www.opensource.org/licenses/mit-license.php
*/

#ifndef _STDAFX_H_
#define _STDAFX_H_

#define FCOLLADA_INTERNAL
#include "FCollada.h"

#endif // _STDAFX_H_
