glext_versions = {
"GL_EXT_gpu_shader4": "3.0",
"GL_NV_conditional_render": "3.0",
"GL_ARB_color_buffer_float": "3.0",
"GL_ARB_depth_buffer_float": "3.0",
"GL_ARB_texture_float": "3.0",
"GL_EXT_packed_float": "3.0",
"GL_EXT_texture_shared_exponent": "3.0",
"GL_EXT_framebuffer_object": "3.0",
"GL_NV_half_float": "3.0",
"GL_ARB_half_float_pixel": "3.0",
"GL_EXT_framebuffer_multisample": "3.0",
"GL_EXT_framebuffer_blit": "3.0",
"GL_EXT_texture_integer": "3.0",
"GL_EXT_texture_array": "3.0",
"GL_EXT_packed_depth_stencil": "3.0",
"GL_EXT_draw_buffers2": "3.0",
"GL_EXT_texture_compression_rgtc": "3.0",
"GL_EXT_transform_feedback": "3.0",
"GL_APPLE_vertex_array_object": "3.0",
"GL_EXT_framebuffer_sRGB": "3.0",

"GL_ARB_pixel_buffer_object": "2.1",
"GL_EXT_texture_sRGB": "2.1",

"GL_ARB_shader_objects": "2.0",
"GL_ARB_vertex_shader": "2.0",
"GL_ARB_fragment_shader": "2.0",
"GL_ARB_shading_language_100": "2.0",
"GL_ARB_draw_buffers": "2.0",
"GL_ARB_texture_non_power_of_two": "2.0",
"GL_ARB_point_sprite": "2.0",
"GL_EXT_blend_equation_separate": "2.0",

"GL_ARB_vertex_buffer_object": "1.5",
"GL_ARB_occlusion_query": "1.5",
"GL_EXT_shadow_funcs": "1.5",

"GL_SGIS_generate_mipmap": "1.4",
"GL_NV_blend_square": "1.4",
"GL_ARB_depth_texture": "1.4",
"GL_ARB_shadow": "1.4",
"GL_EXT_fog_coord": "1.4",
"GL_EXT_multi_draw_arrays": "1.4",
"GL_ARB_point_parameters": "1.4",
"GL_EXT_secondary_color": "1.4",
"GL_EXT_blend_func_separate": "1.4",
"GL_EXT_stencil_wrap": "1.4",
"GL_ARB_texture_env_crossbar": "1.4",
"GL_EXT_texture_lod_bias": "1.4",
"GL_ARB_texture_mirrored_repeat": "1.4",
"GL_ARB_window_pos": "1.4",

"GL_ARB_texture_compression": "1.3",
"GL_ARB_texture_cube_map": "1.3",
"GL_ARB_multisample": "1.3",
"GL_ARB_multitexture": "1.3",
"GL_ARB_transpose_matrix": "1.3",
"GL_ARB_texture_env_add": "1.3",
"GL_ARB_texture_env_combine": "1.3",
"GL_ARB_texture_env_dot3": "1.3",
"GL_ARB_texture_border_clamp": "1.3",

"GL_EXT_texture3D": "1.2",
"GL_EXT_bgra": "1.2",
"GL_EXT_packed_pixels": "1.2",
"GL_EXT_rescale_normal": "1.2",
"GL_EXT_separate_specular_color": "1.2",
"GL_SGIS_texture_edge_clamp": "1.2",
"GL_SGIS_texture_lod": "1.2",
"GL_EXT_draw_range_elements": "1.2",
"GL_EXT_blend_color": "1.2",
"GL_EXT_blend_minmax": "1.2",
"GL_EXT_blend_subtract": "1.2",
}
