#include "error_dialog.h"
#include <windows.h>

#ifdef _WIN32
LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
#pragma code_page(1252)
#endif //_WIN32


/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

#ifdef APSTUDIO_INVOKED
1 TEXTINCLUDE 
BEGIN
    "error_dialog.h\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_DIALOG1 DIALOGEX 0, 0, 328, 260
STYLE DS_SETFONT | DS_FIXEDSYS | WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_VISIBLE | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME
EXSTYLE WS_EX_APPWINDOW
CAPTION "Program Error"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    PUSHBUTTON      "&Continue",IDC_CONTINUE,7,239,50,14
    EDITTEXT        IDC_EDIT1,7,7,314,204,ES_MULTILINE | ES_READONLY | ES_WANTRETURN | WS_VSCROLL
    PUSHBUTTON      "Copy",IDC_COPY,271,212,50,14
    PUSHBUTTON      "&Suppress",IDC_SUPPRESS,59,239,50,14
    PUSHBUTTON      "&Break",IDC_BREAK,111,239,50,14
    PUSHBUTTON      "E&xit",IDC_EXIT,163,239,50,14
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO 
BEGIN
    IDD_DIALOG1, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 321
        TOPMARGIN, 7
        BOTTOMMARGIN, 253
    END
END
#endif    // APSTUDIO_INVOKED

/////////////////////////////////////////////////////////////////////////////
