/*
	Copyright (C) 2005-2007 Feeling Software Inc.
	Portions of the code are:
	Copyright (C) 2005-2007 Sony Computer Entertainment America
	
	MIT License: http://www.opensource.org/licenses/mit-license.php
*/
/**
	@file FCDGeometryNURBSSurface.h
	This file contains the FCDGeometryNURBSSurface class.
	The FCDGeometryNURBSSurface class hold the information for one COLLADA NURBS surface.

	Note: The NURBS surfaces are part of the Premium Build only.
*/

