/*
	Copyright (C) 2005-2007 Feeling Software Inc.
	Portions of the code are:
	Copyright (C) 2005-2007 Sony Computer Entertainment America
	
	MIT License: http://www.opensource.org/licenses/mit-license.php
*/

#ifndef _STD_AFX_PCH_H_
#define _STD_AFX_PCH_H_

#define FCOLLADA_INTERNAL
#include "FCollada.h"

#include "FAXColladaParser.h"
#include "FAXColladaWriter.h"

#include "FUtils/FUXmlDocument.h"
#include "FUtils/FUFileManager.h"
#include "FUtils/FUUniqueStringMap.h"

using namespace FUDaeParser;
using namespace FUDaeWriter;

#include "FAXStructures.h"

#endif // _STD_AFX_PCH_H_

